#define USE_CURSES 1
#define SMES 0

#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sgtty.h>
#include <sys/param.h>
#include <time.h>
#if USE_CURSES
#include <curses.h>
#endif
#include "bp.h"

char title[] = "tipbp 1.00";

int infd = 0, outfd = 1;
int interactive = 0;
int userabort;
char s128[128];
char esci_response[] = "#XXX,PB";

extern int PortBps;
extern int BP_Auto_Resume;
extern int BP_Use_File_Size;
extern long BP_File_Size;
int terminated;


#if !USE_CURSES
int displen = 0;
struct _STVALUE {
    char *fmt;
    long value;
    int flags;
} stval[STNumVal] = {
    {"Message: %s\n",0,1},
    {"Direction: %s\n",0,1},
    {"Type: %s\n",0,1},
    {"File: %s\n",0,1},

    {"CSent(%ld) ",0,4},
    {"PSent(%ld) ",0,4},
    {"ESent(%ld) ",0,1},
    {"CRead(%ld) ",0,4},
    {"PRead(%ld) ",0,4},
    {"ERead(%ld) ",0,1},
    {"RemSize(%ld) ",0,4},
    {"Size(%ld) ",0,4},
    {"Data(%ld) ",0,1},
    {"CRate(%ld) ",0,4},
    {"Rate(%ld) ",0,0},
    {"Elapsed(%ld) ",0,0},
    {"RemTime(%ld) ",0,0}
};
#endif

#if USE_CURSES
WINDOW *win = NULL;

struct _STLOC {
    int row, col;
    int w, f;
} stloc[] = {
/* for ST_Display_String */
{12,3, 55, 1}, /* STMsg */
{5,14, 12, 1}, /* STUpDow */
{5,34, 6, 1}, /* STType */
{2,9, 48, 1}, /* STFile */
/* for ST_Display_Value */
{6,31, 8, 0}, /* STComSent */
{7,31, 8, 0}, /* STPacSent */
{11,24, 4, 4|1}, /* STErrSent */
{6,18, 8, 0}, /* STComRead */
{7,18, 8, 0}, /* STPacRead */
{11,15, 4, 4|1}, /* STErrRead */
{3,49, 8, 0}, /* STUplRem, STDowRem */
{3,31, 8, 0}, /* STUplSize, STDowSize */
{3,14, 8, 1}, /* STDataSent, STDataRead */
{8,31, 5, 0}, /* STComRate */
{8,48, 5, 0}, /* STDataRate */
{10,21, 8, 2}, /* STElapsed */
{10,39, 8, 2}, /* STRemTime */

{8,14, 5, 0}, /* baud rate */
{9,21, 8, 2}, /* started time */
{10,3, 8, 2}, /* now time */
{13,3, 55, 0}, /* yes/no mes */
};

char *wintmpl[] = {
/*00000000001111111111222222222233333333334444444444555555555 */
/*01234567890123456789012345678901234567890123456789012345678 */
/*.----------------------------------------------------------. */
  "                                                          ",	/* 1 */
  "  File: ________________________________________________  ",	/* 2 */
  "  position:  ________ length: ________ remains: ________  ",	/* 3 */
  "                                                          ",	/* 4 */
  "  Direction: ____________  Mode: ______                   ",	/* 5 */
  "  Comm I/O:   rx ________  tx ________ bytes              ",	/* 6 */
  "  Packet I/O: rx ________  tx ________ packets            ", /* 7 */
  "  Baud rate: _____ Comm rate: _____ Data rate: _____      ", /* 8 */
  "  Time:    started: __:__:__                              ", /* 9 */
  "  __:__:__ elapsed: __:__:__ remains: __:__:__            ", /* 10 */
  "  Errors:  rx ____  tx ____                               ", /* 11 */
  "                                                          ",	/* 12 */
  "                                                          ",	/* 13 */
/*`----------------------------------------------------------' */
(char *)0
};

#define WinRows	15
#define WinCols	60
#define WinTopY	2
#define WinLeftX	8
#endif

cancelsig(sig)
int sig;
{
    userabort = 1;
}

abortsig(sig)
int sig;
{
#if USE_CURSES
    if (win != NULL)
	endwin(win);
#endif
    exit(128 + sig);
}

setsignals()
{
    signal(SIGINT, cancelsig);
    signal(SIGTERM, cancelsig);
    signal(SIGQUIT, cancelsig);
}

dflsignals()
{
    signal(SIGINT, abortsig);
    signal(SIGTERM, abortsig);
    signal(SIGQUIT, abortsig);
}

#if USE_CURSES
char *getnow()
{
    long now;
    struct tm *lt;
    static char buff[16];

    now = time((long *)0);
    lt = localtime(&now);
    sprintf(buff, "%02d:%02d:%02d", lt->tm_hour, lt->tm_min, lt->tm_sec);
    return buff;
}

erasearea(win, r, c, l)
WINDOW *win;
int r, c, l;
{
    wmove(win, r, c);
    while (l-- > 0)
	waddch(win, ' ');
    wmove(win, r, c);
}
#endif

/* send a char asynchronous */
Async_Send(ch)
char ch;
{
    write(outfd, &ch, 1);
}

/* check and read char
 * return 1:	ch <- char
 * return 0:	ch undefined
 */
Async_BuffeR_Check(ch)
unsigned char *ch;
{
    static int nrest = 0;
    static char rbuff[BUFSIZ], *rbp;

    if (nrest > 0){
	--nrest;
	*ch = *rbp++;
	return 1;
    }
    ioctl(infd, FIONREAD, &nrest);
    if (nrest <= 0)
	return 0;
    nrest = read(infd, rbuff, sizeof rbuff);
    if (nrest <= 0)
	return 0;
    --nrest;
    rbp = rbuff;
    *ch = *rbp++;
    return 1;
}

/* check user abort
 * return !0 if abort
 */
ST_Check_Abort()
{
    int rval;

    rval = userabort;
    userabort = 0;
    return rval;
}

ST_Display_String(type, mes)
int type;
char *mes;
{
#if USE_CURSES
    struct _STLOC *sl = &stloc[type];

    erasearea(win, sl->row, sl->col, sl->w);
    if (mes != NULL) {
	char buff[80];

	strncpy(buff, mes, sl->w);
	buff[sl->w] = '\0';
	if (sl->f & 4)
	    wstandout(win);
	waddstr(win, buff);
	if (sl->f & 4)
	    wstandend(win);
    }
    if (sl->f & 1) {
	ST_Display_String(STNowTime, getnow());
	wrefresh(win);
    }
#else
    if (displen) {
	putchar('\n');
	displen = 0;
    }
    printf(stval[type].fmt, mes);
#endif
}

ST_Display_Value(type, val)
int type;
long val;
{
#if USE_CURSES
    struct _STLOC *sl = &stloc[type];
    char buff[16];

    if (sl->f & 2)
	sprintf(buff, "%02d:%02d:%02d", (val/3600)%99, (val/60)%60, val%60);
    else
	sprintf(buff, "%*ld", sl->w, val);
    ST_Display_String(type, buff);
#else
    register struct _STVALUE *sp = &stval[type];

    sp->value = val;
    if (!(sp->flags & 1))
	return;
    if (sp->flags & 2) {
	if (displen) {
	    putchar('\n');
	    displen = 0;
	}
	printf(sp->fmt, sp->value);
    } else {
	int i;
	char buff[BUFSIZ], *p;

	p = buff;
	*p = '\0';
	for (i = STSttVal; i < STNumVal; ++i)
	    if (!(stval[i].flags & (4|2)) && stval[i].value != 0) {
		sprintf(p, stval[i].fmt, stval[i].value);
		p += strlen(p);
	    }
	while (p > buff && *(p - 1) == ' ')
	    *--p = '\0';
	if (buff[0] != '\0') {
	    printf("\r%s", buff);
	    i = p - buff;
	    while (i < displen) {
		putchar(' ');
		++i;
	    }
	    displen = p - buff;
	    fflush(stdout);
	}
    }
#endif
}

/* ask yes or no
 * return *yn = 'Y' or *yn = 'N'
 */
ST_Yes_or_No(mes, yn)
char *mes;
char *yn;
{
#if USE_CURSES
    int ans;

    ST_Display_String(STNowTime, getnow());
    ST_Display_String(STYesNoMes, mes);
    waddstr(win, " : ");
    wrefresh(win);
    for (;;) {
	ans = wgetch(win);
	if (ans == 'y' || ans == 'Y') {
	    *yn = 'Y';
	    break;
	} else if (ans == 'n' || ans == 'N') {
	    *yn = 'N';
	    break;
	}
    }
    waddch(win, ans);
    wrefresh(win);
    ST_Display_String(STYesNoMes, NULL);
    ST_Display_String(STNowTime, getnow());
    wrefresh(win);
#else
    char buff[BUFSIZ];

    if (displen) {
	putchar('\n');
	displen = 0;
    }
    printf("%s (yes or no) : ", mes);
    *yn = 'Y';
    while (gets(buff)) {
	if (!strcmp(buff, "yes"))
	    return;
	else if (!strcmp(buff, "no")) {
	    *yn = 'N';
	    return;
	}
    }
#endif
}

ST_Initialize()
{
    userabort = 0;
    setsignals();
}

ST_Terminate()
{
    dflsignals();
    terminated = 1;
}

setscreen()
{
#if USE_CURSES
    struct sgttyb ttyb;
    int i;
    long br;
    char wdbuff[MAXPATHLEN];

    ioctl(infd, TIOCGETP, &ttyb);
    switch (ttyb.sg_ispeed) {
    case B110: PortBps = 0; br = 110L; break;
    case B150: PortBps = 1; br = 150L; break;
    case B300: PortBps = 2; br = 300L; break;
    case B600: PortBps = 3; br = 600L; break;
    case B1200: PortBps = 4; br = 1200L; break;
    case B1800: PortBps = 5; br = 1800L; break;
    case B2400: PortBps = 6; br = 2400L; break;
    case B4800: PortBps = 7; br = 4800L; break;
    case B9600: PortBps = 8; br = 9600L; break;
    case B19200:
    default: PortBps = 9; br = 19200L; break;
    }
    getwd(wdbuff);

    initscr();
    crmode();
    noecho();
    nonl();
    clear();
    win = newwin(WinRows, WinCols, WinTopY, WinLeftX);
    box(win, '|', '-');
    wmove(win, 0, 0); waddch(win, '.');
    wmove(win, win->_maxy - 1, 0); waddch(win, '`');
    wmove(win, win->_maxy - 1, win->_maxx - 1); waddch(win, '\'');
    wmove(win, 0, win->_maxx - 1); waddch(win, '.');
    wmove(win, 0, 2);
    wstandout(win);
    waddch(win, '[');
    waddch(win, ' ');
    waddstr(win, title);
    waddch(win, ' ');
    waddch(win, ']');
    wstandend(win);
    waddch(win, '-');
    waddch(win, '-');
    waddch(win, ' ');
    wdbuff[WinCols - 2 - 7 - strlen(title)] = '\0';
    waddstr(win, wdbuff);
    waddch(win, ' ');
    for (i = 0; wintmpl[i] != (char *)0; ++i) {
	wmove(win, i + 1, 1);
	waddstr(win, wintmpl[i]);
    }
    ST_Display_String(STStartedTime, getnow());
    ST_Display_Value(STBaudRate, br);

    wrefresh(win);
#else
    displen = 0;
#endif
}

unsetscreen()
{
#if USE_CURSES
    endwin();
    win = NULL;
#else
    if (displen)
	printf("\n");
#endif
}

interactivemode()
{
    int prevc = -1;
    unsigned char c;

    setscreen();
    terminated = 0;
    while (!terminated) {
	while (!Async_BuffeR_Check(&c))
	    ;
	switch (c) {
	case 'I':
	    if (prevc == ESC) {
		BP_Term_ESC_I(esci_response);
#if !USE_CURSES&&SMES
		printf("Term ESC I\n");
#endif
		prevc = -1;
	    } else
		prevc = c;
	    break;
	case ENQ:
	    BP_Term_ENQ();
#if !USE_CURSES&&SMES
	    printf("Term ENQ\n");
#endif
	    prevc = -1;
	    break;
	case DLE:
#if !USE_CURSES&&SMES
	    printf("DLE Seen\n");
#endif
	    BP_DLE_Seen();
	    prevc = -1;
	    break;
	default:
	    prevc = c;
	    break;
	}
    }
    unsetscreen();
}

usage(mes)
char *mes;
{
    fprintf(stderr, "%s\n", mes);
    exit(255);
}

main(argc, argv)
int argc;
char **argv;
{
    char *p;
    long atol();
    int bplus = 0;

    dflsignals();
    while (--argc)
	if (*(p = *++argv) == '-')
	    while (*++p)
		switch (*p) {
		case 'Q':
		    bplus = 1;
		    break;
		case 'E':
		    bplus = 2;
		    break;
		case 'D':
		    bplus =3;
		    break;
		case 'I':
		    interactive = 1;
		    bplus = 4;
		    break;
		case '.':
		    if (--argc < 1)
			usage("no iofd after -.");
		    infd = outfd = atoi(*++argv);
		    break;
		case 'S':
		    if (--argc < 1)
			usage("no filesize after -S");
		    BP_Use_File_Size = 1;
		    BP_File_Size = atol(*++argv);
		    break;
		case 'R':
		    BP_Auto_Resume = 1;
		    break;
		default:
		    sprintf(s128, "Unknown switch -%c", *p);
		    usage(s128);
		    break;
		}
	else
	    usage("don't specify other arguments");
    if (bplus == 0 || infd != outfd) {
	printf("can only be run by tip\n");
	exit(255);
    }

#if !USE_CURSES
    printf("%s\n", title);
#endif
    userabort = 0;
    switch (bplus) {
    case 1:
	BP_Term_ENQ();
	break;
    case 2:
	BP_Term_ESC_I(esci_response);
	break;
    case 3:
	setscreen();
	BP_DLE_Seen();
	unsetscreen();
	break;
    case 4:
	interactivemode();
	break;
    }

    return 0;
}
